import os
import re
from typing import Tuple, Union, List
import mimetypes

from qase.commons.models import Relation
from qase.commons.models.config.framework import Trace, Video
from qase.commons.models.relation import SuiteData
from qase.commons.models.result import Result, Field
from qase.commons.models.attachment import Attachment
from qase.commons.models.runtime import Runtime

from qase.commons import QaseUtils

from filelock import FileLock
from itertools import chain, islice

PYTEST_TO_QASE_STATUS = {
    "PASSED": 'passed',
    "FAILED": 'failed',
    "SKIPPED": 'skipped',
    "BLOCKED": 'blocked',
    "BROKEN": 'invalid'
}

try:
    from xdist import is_xdist_controller
except ImportError:
    def is_xdist_controller(*args, **kwargs):
        return True

try:
    import pytest
except ImportError:
    raise ImportError("pytest is not installed")


class PluginNotInitializedException(Exception):
    pass


class QasePytestPlugin:
    run = None
    meta_run_file = "src.run"

    def __init__(
            self,
            reporter
    ):
        self.runtime = Runtime()
        self.reporter = reporter
        self.config = reporter.config
        self.run_id = None
        self.execution_plan = reporter.get_execution_plan()
        self._current_item = None
        self.ignore = None

        self.reporter.setup_profilers(runtime=self.runtime)

    def start_step(self, step):
        self.runtime.add_step(step)

    def finish_step(self, id, exception=None):
        status = PYTEST_TO_QASE_STATUS['PASSED']
        if exception:
            status = PYTEST_TO_QASE_STATUS['FAILED']

        self.runtime.finish_step(id, status=status)

    @staticmethod
    def drop_run_id():
        if os.path.exists(QasePytestPlugin.meta_run_file):
            os.remove(QasePytestPlugin.meta_run_file)

    def pytest_collection_modifyitems(self, session, config, items):
        """
        Modify collected test items to inject our parameter capturing logic.
        """
        for item in items:
            grouped_params = []

            # Extract single and grouped params from the item's markers
            for mark in item.iter_markers():
                if mark.name == 'parametrize':
                    if len(mark.args) != 0:
                        param_name, values = mark.args
                        if ',' in param_name:
                            grouped_params.append(
                                [item.strip() for item in param_name.split(',')])
                    else:
                        param_name = mark.kwargs.get('argnames')
                        if ',' in param_name:
                            grouped_params.append(
                                [item.strip() for item in param_name.split(',')])

            # Attach the captured params to the test item
            item._grouped_params = grouped_params

        # Filter test cases based on ids
        if self.execution_plan:
            new_items = []
            for item in items:
                if item.get_closest_marker('qase_id'):
                    ids = QasePytestPlugin._get_qase_ids(item)
                    if any(id in self.execution_plan for id in ids):
                        new_items.append(item)

            items[:] = new_items

    def pytest_sessionstart(self, session):
        if is_xdist_controller(session):
            self.run_id = self.reporter.start_run()
            with FileLock("qase.lock"):
                if self.run_id:
                    with open(self.meta_run_file, "w") as lock_file:
                        lock_file.write(str(self.run_id))
        else:
            self.load_run_from_lock()

    def pytest_sessionfinish(self, session, exitstatus):
        if is_xdist_controller(session):
            QasePytestPlugin.drop_run_id()
        else:
            self.reporter.complete_worker()

        if not os.path.exists(QasePytestPlugin.meta_run_file):
            self.reporter.complete_run()

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_protocol(self, item):
        self.ignore = True if item.get_closest_marker("qase_ignore") else False

        if not self.ignore:
            self.reporter.enable_profilers()
            self.start_pytest_item(item)
            self._current_item = item
            yield
            self._current_item = None
            self.reporter.disable_profilers()
        else:
            yield

    @pytest.hookimpl
    def pytest_runtest_logfinish(self):
        if not self.ignore and self._current_item:
            self.finish_pytest_item()
            self.start_pytest_item(self._current_item)

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_makereport(self, item, call):
        if self.ignore:
            yield
            return

        report = (yield).get_result()

        # Skip processing if not in the relevant phase
        if call.when not in ["setup", "call"]:
            return

        # Process test result and update status
        self._process_test_result(report, call, item)

        # Add logs if configured and in call phase
        if self.reporter.config.framework.pytest.capture_logs and call.when == "call":
            self._attach_logs(report)

        # Handle xfail message in call phase
        if (call.when == "call" and
                self.__is_xfail_mark(report) and
                self.runtime.result.execution.status != PYTEST_TO_QASE_STATUS['PASSED']):
            self.runtime.result.add_message(report.wasxfail)

        # Handle skip message in setup phase
        self._handle_skip_message_in_setup(report, call)

        # Handle Playwright artifacts if applicable
        if call.when == "call":
            self._process_playwright_artifacts(item)

    def _process_test_result(self, report, call, item):
        """Process test results and set appropriate status."""
        # Add stacktrace if available
        if report.longrepr:
            self.runtime.result.execution.stacktrace = report.longreprtext

        # Handle failed tests
        if report.failed:
            self._handle_failed_test(call)
        # Handle skipped tests
        elif report.skipped:
            self._handle_skipped_test(report, call)
        # Handle passed tests
        else:
            self._handle_passed_test(report, call)

    def _handle_failed_test(self, call):
        """Handle failed test case and set appropriate status."""
        is_assertion_error = False
        error_message = "Test failed"
        
        if call.excinfo is not None:
            is_assertion_error = call.excinfo.typename == "AssertionError"
            error_message = call.excinfo.exconly()
        
        status = PYTEST_TO_QASE_STATUS['FAILED'] if is_assertion_error else PYTEST_TO_QASE_STATUS['BROKEN']
        self._set_result_status(status)
        self.runtime.result.add_message(error_message)

    def _handle_skipped_test(self, report, call):
        """Handle skipped test case and set appropriate status."""
        xfail_status = self._get_xfail_status(report, call)
        if xfail_status:
            self._set_result_status(xfail_status)
        elif self.runtime.result.execution.status in (None, PYTEST_TO_QASE_STATUS['PASSED']):
            self._set_result_status(PYTEST_TO_QASE_STATUS['SKIPPED'])

    def _handle_passed_test(self, report, call):
        """Handle passed test case and set appropriate status."""
        xfail_status = self._get_xfail_status(report, call)
        if xfail_status:
            self._set_result_status(xfail_status)
        elif self.runtime.result.execution.status is None:
            self._set_result_status(PYTEST_TO_QASE_STATUS['PASSED'])

    def _get_xfail_status(self, report, call):
        """Get xfail status if applicable."""
        if self.__is_xfail_mark(report) and call.when == "call":
            status = self.config.framework.pytest.xfail_status
            return status.xfail if report.skipped else status.xpass
        return None

    def _set_result_status(self, status):
        """Set test result execution status."""
        self.runtime.result.execution.status = status

    def _handle_skip_message_in_setup(self, report, call):
        """Handle skip message in setup phase with proper longrepr handling."""
        if (call.when == "setup" and
                report.longrepr and
                self.runtime.result.execution.status == PYTEST_TO_QASE_STATUS['SKIPPED']):
            # Handle different types of longrepr
            skip_message = None
            if hasattr(report.longrepr, '__getitem__'):
                # Handle tuple/list format (old pytest versions)
                try:
                    skip_message = report.longrepr[2].split(":")[1]
                except (IndexError, AttributeError):
                    pass
            elif hasattr(report.longrepr, 'reprcrash'):
                # Handle ExceptionChainRepr format (new pytest versions)
                try:
                    skip_message = str(report.longrepr.reprcrash.message)
                except AttributeError:
                    pass
            
            if skip_message:
                self.runtime.result.add_message(skip_message)

    def _attach_logs(self, report):
        """Attach logs to the test result."""
        logs = [
            (report.caplog, "text/plain", "log.txt"),
            (report.capstdout, "text/plain", "stdout.txt"),
            (report.capstderr, "text/plain", "stderr.txt"),
        ]
        for content, mime_type, filename in logs:
            if content:
                self.add_attachments((content, mime_type, filename))

    def _process_playwright_artifacts(self, item):
        """Process and attach Playwright artifacts if available."""
        if not (hasattr(item, 'funcargs') and 'page' in item.funcargs):
            return

        page = item.funcargs['page']
        if not hasattr(page, 'video'):
            return

        folder_name = self.__build_folder_name(item)
        output_dir = self.config.framework.playwright.output_dir
        base_path = os.path.join(os.getcwd(), output_dir, folder_name)

        # Attach video if enabled
        if self.config.framework.playwright.video != Video.off:
            video_path = os.path.join(base_path, "video.webm")
            self.add_attachments(video_path)

        # Attach trace if enabled
        if self.config.framework.playwright.trace != Trace.off:
            trace_path = os.path.join(base_path, "trace.zip")
            self.add_attachments(trace_path)

    def start_pytest_item(self, item):
        self.runtime.result = Result(
            title=self._get_title(item),
            signature='',
        )
        self._set_fields(item)
        self._set_author(item)
        self._set_muted(item)
        self._set_testops_ids(item)
        self._set_params(item)
        self._set_suite(item)
        self._set_relations(item)
        self._get_signature(item)

    def finish_pytest_item(self):
        self.runtime.result.execution.complete()
        self.runtime.result.add_steps(
            [step for key, step in self.runtime.steps.items()])
        self.reporter.add_result(self.runtime.result)

        self.runtime.clear()

    def add_attachments(
            self, *files: Union[str, Tuple[str, str], Tuple[bytes, str, str]]
    ):
        for file in files:
            filename = None
            content = None
            file_path = None

            if isinstance(file, tuple):
                if len(file) == 2:
                    file_path, mime = file
                    filename = QaseUtils.get_filename(file_path)
                else:
                    content, mime, filename = file
            elif isinstance(file, str):
                file_path = file
                mime = mimetypes.guess_type(file)[0]
                filename = QaseUtils.get_filename(file_path)

            attachment = Attachment(
                file_name=filename, content=content, mime_type=mime, file_path=file_path)
            self.runtime.add_attachment(attachment)

    def add_param(self, name: str, value: str):
        """
        Add a parameter to the current test result.
        """
        if not self.runtime.result:
            return
        self.runtime.result.add_param(name, value)

    def load_run_from_lock(self):
        if os.path.exists(QasePytestPlugin.meta_run_file):
            with open(QasePytestPlugin.meta_run_file, "r") as lock_file:
                try:
                    self.run_id = str(lock_file.read())
                    self.reporter.set_run_id(self.run_id)
                except ValueError:
                    pass
        else:
            self.run_id = self.reporter.start_run()

    def _get_title(self, item):
        title = None
        try:
            title = item.get_closest_marker("qase_title").kwargs.get("title")
        except:
            pass

        if not title:
            title = item.originalname

        return str(title)

    def _get_signature(self, item):
        self.runtime.result.signature = QaseUtils.get_signature(
            self.runtime.result.testops_ids, item.nodeid.split("/"), self.runtime.result.params)

    def _set_relations(self, item) -> None:
        # TODO: Add support for relations
        pass

    def _set_fields(self, item) -> None:
        # Legacy fields support
        for name in ["description", "preconditions", "postconditions", "layer", "severity", "priority", "suite"]:
            try:
                self.runtime.result.add_field(
                    Field(name, item.get_closest_marker("qase_" + name).kwargs.get(name)))
            except:
                pass

        try:
            fields = item.get_closest_marker(
                "qase_fields").kwargs.get("fields")
            for name, field in fields:
                self.runtime.result.add_field(Field(name, field))
        except:
            pass

    def _set_author(self, item) -> None:
        try:
            author = str(item.get_closest_marker(
                "qase_author").kwargs.get("author"))
            if author != "None":
                self.runtime.result.add_field(Field("author", author))
        except:
            pass

    def _set_muted(self, item) -> None:
        try:
            muted = True if item.get_closest_marker("qase_muted") else False
            if muted:
                self.runtime.result.add_field(Field("muted", "true"))
        except:
            pass

    def _set_testops_ids(self, item) -> None:
        try:
            self.runtime.result.testops_ids = QasePytestPlugin._get_qase_ids(
                item)
        except:
            pass

    def _set_params(self, item) -> None:
        if hasattr(item, 'callspec'):
            # Get parameters that should be ignored
            ignored_params = set()
            for mark in item.iter_markers():
                if mark.name == 'qase_parametrize_ignore':
                    param_name = mark.kwargs.get('argnames')
                    if param_name:
                        if ',' in param_name:
                            ignored_params.update(
                                [p.strip() for p in param_name.split(',')])
                        else:
                            ignored_params.add(param_name)
                elif mark.name == 'qase_ignore_parameters':
                    parameters = mark.kwargs.get('parameters', [])
                    if parameters:
                        ignored_params.update(parameters)

            # Add grouped parameters that are not ignored
            for group in item._grouped_params:
                if not any(param in ignored_params for param in group):
                    self.runtime.result.add_param_groups(group)

            # Add individual parameters that are not ignored
            ids = item.callspec.id.split("-")
            if len(ids) != len(item.callspec.params.items()):
                for key, val in item.callspec.params.items():
                    if key.startswith("__pytest") or key in ignored_params:
                        continue
                    self.runtime.result.add_param(key, str(val))
            else:
                i = 0
                for key, val in item.callspec.params.items():
                    if key.startswith("__pytest") or key in ignored_params:
                        i += 1
                        continue
                    value = str(ids[i])
                    i += 1
                    self.runtime.result.add_param(key, value)

    def _set_suite(self, item) -> None:
        marker = item.get_closest_marker("qase_suite")
        if marker:
            self.runtime.result.relations = self.__prepare_relations(
                marker.kwargs.get("title").split('.'))
            return
        self._get_suite(item)

    def _get_suite(self, item):
        path, class_name, tail = islice(
            chain(item.nodeid.split('::'), [None], [None]), 3)

        class_name = class_name if tail else None
        file_name, file_path = islice(
            chain(reversed(path.rsplit('/', 1)), [None]), 2)

        module = file_name.split('.')[0]

        if file_path:
            title = file_path + '.' + module
        else:
            title = module

        if class_name:
            title += '.' + class_name

        self.runtime.result.relations = self.__prepare_relations(
            title.split('.'))

    @staticmethod
    def __prepare_relations(suites: []):
        relation = Relation()

        for suite in suites:
            relation.add_suite(SuiteData(title=suite))

        return relation

    @staticmethod
    def __build_folder_name(item):
        path_parts = [
            QasePytestPlugin.__sanitize_path_component(item.location[0])]

        if '.' in item.location[2]:
            path_parts.append(
                QasePytestPlugin.__sanitize_path_component(
                    item.location[2].split('.')[0])
            )

        path_parts.append(
            QasePytestPlugin.__sanitize_path_component(item.originalname)
        )

        if 'browser_name' in item.funcargs:
            path_parts.append(item.funcargs['browser_name'])

        return "-".join(path_parts)

    @staticmethod
    def __sanitize_path_component(component):
        return component.replace(os.sep, "-").replace(".", "-").replace("_", "-").lower()

    @staticmethod
    def __is_xfail_mark(report):
        """Check if the test is marked as xfail."""
        return hasattr(report, 'wasxfail')

    @staticmethod
    def _get_qase_ids(item) -> Union[None, List[int]]:
        ids = item.get_closest_marker("qase_id").kwargs.get("id")
        if ids is None:
            return None
        if isinstance(ids, int):
            return [ids]
        else:
            return [int(i) for i in re.split(r'\s*,\s*', ids)]


class QasePytestPluginSingleton:
    _instance = None

    @staticmethod
    def init(**kwargs):
        if QasePytestPluginSingleton._instance is None:
            QasePytestPluginSingleton._instance = QasePytestPlugin(**kwargs)

    @staticmethod
    def get_instance() -> QasePytestPlugin:
        """ Static access method"""
        if QasePytestPluginSingleton._instance is None:
            raise PluginNotInitializedException("Init plugin first")
        return QasePytestPluginSingleton._instance

    def __init__(self):
        """ Virtually private constructor"""
        raise Exception("Use get_instance()")



from .plugin import QasePytestPlugin, QasePytestPluginSingleton

from qase.commons.reporters import QaseCoreReporter

from .options import QasePytestOptions
from qase.commons.config import ConfigManager


def pytest_addoption(parser):
    group = parser.getgroup("qase")
    QasePytestOptions.addoptions(parser, group)


def pytest_configure(config):
    config = _add_markers(config)
    config.addinivalue_line(
        "markers",
        "qase_parametrize_ignore: mark test to ignore parameters in Qase reports"
    )

    config_manager = setup_config_manager(config)

    QasePytestPluginSingleton.init(reporter=QaseCoreReporter(
        config_manager, 'pytest', 'qase-pytest'))
    config.qase = QasePytestPluginSingleton.get_instance()
    config.pluginmanager.register(
        config.qase,
        name="qase-pytest",
    )


def _add_markers(config):
    config.addinivalue_line(
        "markers", "qase_id: mark test to be associate with Qase TestOps test case")
    config.addinivalue_line("markers", "qase_title: mark test with title")
    config.addinivalue_line(
        "markers", "qase_ignore: skip test from Qase TestOps and Qase Report")
    config.addinivalue_line(
        "markers", "qase_muted: mark test as muted so it will not affect test run status")
    config.addinivalue_line("markers", "qase_author: mark test with author")
    config.addinivalue_line("markers", "qase_fields: mark test with meta data")
    config.addinivalue_line("markers", "qase_suite: mark test with suite")
    config.addinivalue_line(
        "markers", "qase_ignore_parameters: mark test to ignore specific parameters in Qase reports")

    # Legacy markers | These markers are deprecated and will be removed in future versions
    config.addinivalue_line(
        "markers", "qase_description: mark test with description")
    config.addinivalue_line(
        "markers", "qase_preconditions: mark test with preconditions")
    config.addinivalue_line(
        "markers", "qase_postconditions: mark test with postconditions")
    config.addinivalue_line("markers", "qase_layer: mark test with layer")
    config.addinivalue_line(
        "markers", "qase_severity: mark test with severity")
    config.addinivalue_line(
        "markers", "qase_priority: mark test with priority")
    config.addinivalue_line("markers", "qase_tags: mark test with tags")
    return config


def setup_config_manager(config):
    config_manager = ConfigManager()
    for option in config.option.__dict__:
        if option == "output" and config.option.__dict__[option] is not None:
            config_manager.config.framework.playwright.set_output_dir(
                config.option.__dict__[option])

        if option == "video" and config.option.__dict__[option] is not None:
            config_manager.config.framework.playwright.set_video(
                config.option.__dict__[option])

        if option == "tracing" and config.option.__dict__[option] is not None:
            config_manager.config.framework.playwright.set_trace(
                config.option.__dict__[option])

        if option.startswith("qase_"):
            if option == "qase_mode" and config.option.__dict__[option] is not None:
                config_manager.config.set_mode(config.option.__dict__[option])

            if option == "qase_fallback" and config.option.__dict__[option] is not None:
                config_manager.config.set_fallback(
                    config.option.__dict__[option])

            if option == "qase_environment" and config.option.__dict__[option] is not None:
                config_manager.config.set_environment(
                    config.option.__dict__[option])

            if option == "qase_profilers" and config.option.__dict__[option] is not None:
                config_manager.config.set_profilers(
                    config.option.__dict__[option].split(","))

            if option == "qase_root_suite" and config.option.__dict__[option] is not None:
                config_manager.config.set_root_suite(
                    config.option.__dict__[option])

            if option == "qase_debug" and config.option.__dict__[option] is not None:
                config_manager.config.set_debug(config.option.__dict__[option])

            if option == "qase_exclude_params" and config.option.__dict__[option] is not None:
                config_manager.config.set_exclude_params(
                    [param.strip() for param in config.option.__dict__[option].split(',')])

            if option == "qase_execution_plan_path" and config.option.__dict__[option] is not None:
                config_manager.config.execution_plan.set_path(
                    config.option.__dict__[option])

            if option == "qase_testops_project" and config.option.__dict__[option] is not None:
                config_manager.config.testops.set_project(
                    config.option.__dict__[option])

            if option == "qase_testops_api_token" and config.option.__dict__[option] is not None:
                config_manager.config.testops.api.set_token(
                    config.option.__dict__[option])

            if option == "qase_testops_api_host" and config.option.__dict__[option] is not None:
                config_manager.config.testops.api.set_host(
                    config.option.__dict__[option])

            if option == "qase_testops_plan_id" and config.option.__dict__[option] is not None:
                config_manager.config.testops.plan.set_id(
                    int(config.option.__dict__[option]))

            if option == "qase_testops_run_id" and config.option.__dict__[option] is not None:
                config_manager.config.testops.run.set_id(
                    int(config.option.__dict__[option]))

            if option == "qase_testops_run_title" and config.option.__dict__[option] is not None:
                config_manager.config.testops.run.set_title(
                    config.option.__dict__[option])

            if option == "qase_testops_run_description" and config.option.__dict__[option] is not None:
                config_manager.config.testops.run.set_description(
                    config.option.__dict__[option])

            if option == "qase_testops_run_complete" and config.option.__dict__[option] is not None:
                config_manager.config.testops.run.set_complete(
                    config.option.__dict__[option])

            if option == "qase_testops_run_tags" and config.option.__dict__[option] is not None:
                config_manager.config.testops.run.set_tags(
                    [tag.strip() for tag in config.option.__dict__[option].split(',')])

            if option == "qase_testops_defect" and config.option.__dict__[option] is not None:
                config_manager.config.testops.set_defect(
                    config.option.__dict__[option])

            if option == "qase_report_driver" and config.option.__dict__[option] is not None:
                config_manager.config.report.set_driver(
                    config.option.__dict__[option])

            if option == "qase_report_connection_local_path" and config.option.__dict__[option] is not None:
                config_manager.config.report.connection.set_path(
                    config.option.__dict__[option])

            if option == "qase_report_connection_local_format" and config.option.__dict__[option] is not None:
                config_manager.config.report.connection.set_format(
                    config.option.__dict__[option])

            if option == "qase_testops_batch_size" and config.option.__dict__[option] is not None:
                config_manager.config.testops.batch.set_size(
                    int(config.option.__dict__[option]))

            if option == "qase_pytest_capture_logs" and config.option.__dict__[option] is not None:
                config_manager.config.framework.pytest.set_capture_logs(
                    config.option.__dict__[option])

            if option == "qase_pytest_xfail_status" and config.option.__dict__[option] is not None:
                config_manager.config.framework.pytest.xfail_status.set_xfail(
                    config.option.__dict__[option])

            if option == "qase_pytest_xpass_status" and config.option.__dict__[option] is not None:
                config_manager.config.framework.pytest.xfail_status.set_xpass(
                    config.option.__dict__[option])

            if option == "qase_testops_configurations_values" and config.option.__dict__[option] is not None:
                # Parse configurations from CLI parameter
                # Format: "group1=value1,group2=value2"
                config_pairs = config.option.__dict__[option].split(',')
                for pair in config_pairs:
                    if '=' in pair:
                        name, config_value = pair.split('=', 1)
                        config_manager.config.testops.configurations.add_value(name.strip(), config_value.strip())

            if option == "qase_testops_configurations_create_if_not_exists" and config.option.__dict__[option] is not None:
                config_manager.config.testops.configurations.set_create_if_not_exists(
                    config.option.__dict__[option])

            if option == "qase_testops_status_filter" and config.option.__dict__[option] is not None:
                config_manager.config.testops.set_status_filter(
                    [status.strip() for status in config.option.__dict__[option].split(',')])

            if option == "qase_testops_show_public_report_link" and config.option.__dict__[option] is not None:
                config_manager.config.testops.set_show_public_report_link(
                    config.option.__dict__[option])

            if option == "qase_status_mapping" and config.option.__dict__[option] is not None:
                # Parse status mapping from CLI parameter
                # Format: "source1=target1,source2=target2"
                mapping_dict = {}
                pairs = config.option.__dict__[option].split(',')
                for pair in pairs:
                    pair = pair.strip()
                    if pair and '=' in pair:
                        source_status, target_status = pair.split('=', 1)
                        mapping_dict[source_status.strip()] = target_status.strip()
                config_manager.config.set_status_mapping(mapping_dict)

            if option == "qase_testops_run_external_link_type" and config.option.__dict__[option] is not None:
                if not config_manager.config.testops.run.external_link:
                    from qase.commons.models.external_link import ExternalLinkConfig
                    config_manager.config.testops.run.external_link = ExternalLinkConfig()
                config_manager.config.testops.run.external_link.set_type(config.option.__dict__[option])

            if option == "qase_testops_run_external_link_url" and config.option.__dict__[option] is not None:
                if not config_manager.config.testops.run.external_link:
                    from qase.commons.models.external_link import ExternalLinkConfig
                    config_manager.config.testops.run.external_link = ExternalLinkConfig()
                config_manager.config.testops.run.external_link.set_link(config.option.__dict__[option])

            if option == "qase_logging_console" and config.option.__dict__[option] is not None:
                config_manager.config.logging.set_console(config.option.__dict__[option])

            if option == "qase_logging_file" and config.option.__dict__[option] is not None:
                config_manager.config.logging.set_file(config.option.__dict__[option])

    # Update logger with final logging options after all CLI options are processed
    from qase.commons.logger import LoggingOptions
    logging_options = LoggingOptions(
        console=config_manager.config.logging.console if config_manager.config.logging.console is not None else True,
        file=config_manager.config.logging.file if config_manager.config.logging.file is not None else config_manager.config.debug
    )
    config_manager.logger = config_manager.logger.__class__(debug=config_manager.config.debug, logging_options=logging_options)

    return config_manager


def pytest_unconfigure(config):
    qase = getattr(config, "src", None)
    if qase:
        del config.qase
        config.pluginmanager.unregister(qase)



import uuid
from typing import Sequence, Tuple, Union, List, Any, Callable

import pytest

from .plugin import PluginNotInitializedException, QasePytestPluginSingleton
from qase.commons.models.step import Step, StepTextData, StepType

from .context_manager import contextdecorator


class qase:
    """Class with decorators for pytest"""

    @staticmethod
    def id(id: Union[int, List[int]]):
        """
        Define a persisent connection to Qase TestOps test case.

        >>> @qase.id(1)
        >>> def test_example():
        >>>     pass

        >>> @qase.id([2,3])
        >>> def test_example():
        >>>     pass

        :param id: int or list of int id of test case
        :return: pytest.mark instance
        """

        if isinstance(id, list) and all(isinstance(item, int) for item in id):
            return pytest.mark.qase_id(id=','.join(str(i) for i in id))

        return pytest.mark.qase_id(id=id)

    @staticmethod
    def title(title):
        """
        >>> @qase.title("Sign up")
        >>> def test_example():
        >>>     pass

        :param title: a string with test name
        :return: pytest.mark instance
        """
        return pytest.mark.qase_title(title=title)

    @staticmethod
    def fields(*fields: Tuple[str, str]):
        """
        >>> @qase.fields(
                ("priority", "high"),
                ("layer", "API")
                ("custom_field", "custom_value")
            )
        >>> def test_example():
        >>>     pass

        :param name: a string with field name
        :param value: a string with field value
        :return: pytest.mark instance
        """
        return pytest.mark.qase_fields(fields=fields)

    @staticmethod
    def suite(title: str, description: str = None):
        """
        >>> @qase.suite("Sign up")
        >>> def test_example():
        >>>     pass

        :param title: a string with suite name. You can use dot notation to create nested suites.
        :param description: a string with suite description
        :return: pytest.mark instance
        """
        return pytest.mark.qase_suite(title=title, description=description)

    @staticmethod
    def author(author):
        """
        >>> @qase.author("John Doe")
        >>> def test_example():
        >>>     pass

        :param author: a string with test author
        :return: pytest.mark instance
        """
        return pytest.mark.qase_author(author=author)

    @staticmethod
    def description(description):
        """
        >>> @qase.description("Sign up user using login and password")
        >>> def test_example():
        >>>     pass

        :param description: a string with test full description. Markdown is supported.
        :return: pytest.mark instance
        """
        return pytest.mark.qase_description(description=description)

    @staticmethod
    def preconditions(preconditions):
        """
        >>> @qase.preconditions("Sign up user using login and password")
        >>> def test_example():
        >>>     pass

        :param preconditions: a string with test preconditions. Markdown is supported.
        :return: pytest.mark instance
        """
        return pytest.mark.qase_preconditions(preconditions=preconditions)

    @staticmethod
    def postconditions(postconditions):
        """
        >>> @qase.postconditions("Sign up user using login and password")
        >>> def test_example():
        >>>     pass

        :param postconditions: a string with test postconditions. Markdown is supported.
        :return: pytest.mark instance
        """
        return pytest.mark.qase_postconditions(postconditions=postconditions)

    @staticmethod
    def severity(severity):
        """
        >>> @qase.severity("critical")
        >>> def test_example():
        >>>     pass

        :param severity: a string with test severity. Default values: blocker, critical, major, normal, minor, trivial.
        :return: pytest.mark instance
        """
        return pytest.mark.qase_severity(severity=severity)

    @staticmethod
    def priority(priority):
        """
        >>> @qase.priority("Critical")
        >>> def test_example():
        >>>     pass

        :param priority: a string with test priority. Default values: low, medium, high.
        :return: pytest.mark instance
        """
        return pytest.mark.qase_priority(priority=priority)

    @staticmethod
    def layer(layer):
        """
        >>> @qase.layer("E2E")
        >>> def test_example():
        >>>     pass

        :param layer: a string with test layer. Default values: E2E, API, Unit.
        :return: pytest.mark instance
        """
        return pytest.mark.qase_layer(layer=layer)

    @staticmethod
    def ignore():
        """
        >>> @qase.ignore()
        >>> def test_example():
        >>>     pass

        :return: pytest.mark instance
        """
        return pytest.mark.qase_ignore()

    @staticmethod
    def muted():
        """
        >>> @qase.muted()
        >>> def test_example():
        >>>     pass

        :return: pytest.mark instance
        """
        return pytest.mark.qase_muted()

    @staticmethod
    def attach(*files: Union[str, Tuple[str, str], Tuple[bytes, str, str]]):
        """
        Attach files to test results

        `files` could be:
            - str - only `filepath`
            - str, str - `filepath` and `mime-type` for it
            - bytes, str, str - `source` data, `mime-type` and `filename`

        >>> from src.client.models import MimeTypes
        ... qase.attach(
        ...     (driver.get_screenshot_as_png(), MimeTypes.PNG, "page.png")
        ... )
        """
        try:
            plugin = QasePytestPluginSingleton.get_instance()
            plugin.add_attachments(*files)
        except Exception:
            print("Failed to add attachments")
            pass

    @staticmethod
    def param(name: str, value: str):
        """
        Add a parameter to the current test.

        `name` and `value` are strings.

        >>>
        ... qase.param("param_name", "param_value")
        """
        try:
            plugin = QasePytestPluginSingleton.get_instance()
            plugin.add_param(name=name, value=value)
        except Exception:
            print("Failed to add parameter")
            pass

    @staticmethod
    @contextdecorator
    def step(title, expected=None):
        """
        Step context/decorator

        Usage:

        >>> @qase.step("First step")
        ... def first_step():
        ...     print("smthng")

        >>> with qase.step("Second step"):
        ...     print("smthng")
        """
        plugin = None
        id = str(uuid.uuid4())
        try:
            plugin = QasePytestPluginSingleton.get_instance()
            step = Step(
                step_type=StepType.TEXT,
                id=id,
                data=StepTextData(
                    action=title,
                    expected_result=expected if expected else None,
                )
            )
            plugin.start_step(step)
            yield
            plugin.finish_step(id=step.id)
        except PluginNotInitializedException:
            yield
        except AttributeError:
            yield
        except Exception as e:
            plugin.finish_step(id=id, exception=e)
            raise e

    @staticmethod
    def parametrize_ignore(
        argnames: Union[str, Sequence[str]],
        argvalues: List[Any],
        ids: List[str] = None,
        indirect: Union[bool, List[str]] = False
    ):
        """
        Extended version of pytest.mark.parametrize that prevents parameter collection for Qase reports.
        The test will still be reported, but without parameter values.

        >>> @qase.parametrize_ignore(
        ...     "test_input,expected",
        ...     [("3+5", 8), ("2+4", 6), ("6*9", 42)],
        ...     ids=["add_3_5", "add_2_4", "multiply_6_9"],
        ...     indirect=True
        ... )
        >>> def test_eval(test_input, expected):
        ...     assert eval(test_input) == expected

        :param argnames: A comma-separated string or list of strings denoting one or more argument names
        :param argvalues: The list of argvalues to use for the parametrized test
        :param ids: Optional list of test IDs
        :param indirect: If True, argvalues will be passed as keyword arguments to the test function.
            If a list, it should contain the names of arguments that should be treated as indirect parameters.
        :return: pytest.mark instance
        """
        def decorator(func):
            func = pytest.mark.parametrize(
                argnames, argvalues, ids=ids, indirect=indirect)(func)
            func = pytest.mark.qase_parametrize_ignore(argnames=argnames)(func)
            return func

        return decorator

    @staticmethod
    def ignore_parameters(*parameters: str):
        """
        Mark specific parameters to be ignored in Qase reports.
        This decorator can be used in combination with existing parametrize decorators.

        >>> @pytest.mark.parametrize("browser", ["chrome", "firefox"])
        ... @pytest.mark.parametrize("user", ["user1", "user2"])
        ... @qase.ignore_parameters("user", "browser")
        ... def test_login(browser, user):
        ...     # Both browser and user parameters will be ignored in Qase reports
        ...     pass

        >>> @pytest.mark.parametrize("browser", ["chrome", "firefox"])
        ... @pytest.mark.parametrize("user", ["user1", "user2"])
        ... @qase.ignore_parameters("user")
        ... def test_login(browser, user):
        ...     # Only user parameter will be ignored, browser will be included
        ...     pass

        :param parameters: Variable number of parameter names to ignore
        :return: pytest.mark instance
        """
        return pytest.mark.qase_ignore_parameters(parameters=parameters)



# -*- coding: utf-8 -*-
import functools
from contextlib import ContextDecorator as PyContextDecorator
from contextlib import _GeneratorContextManager as GeneratorContextManager


class ContextManager(GeneratorContextManager, PyContextDecorator):
    """Pass in a generator to the initializer and the resultant object
    is both a decorator closure and context manager
    """

    def __init__(self, func, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}

        super().__init__(func, args, kwargs)


def contextdecorator(func):
    """Similar to contextlib.contextmanager except the decorated generator
    can be used as a decorator with optional arguments.
    """

    @functools.wraps(func)
    def helper(*args, **kwargs):
        is_decorating = len(args) == 1 and callable(args[0])

        if is_decorating:
            new_func = args[0]

            @functools.wraps(new_func)
            def new_helper(*args, **kwargs):
                instance = ContextManager(func)
                return instance(new_func)(*args, **kwargs)

            return new_helper
        return ContextManager(func, args, kwargs)

    return helper



class QasePytestOptions:

    @staticmethod
    def addoptions(parser, group):
        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-mode",
            dest="qase_mode",
            help="Define Qase reporter mode: `off`, `report` or `testops`"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-fallback",
            dest="qase_fallback",
            help="Define Qase reporter fallback mode: `off`, `report` or `testops`"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-environment",
            dest="qase_environment",
            help="Define environment slug from TestOps"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-debug",
            dest="qase_debug",
            type="bool",
            help="Enable debug mode"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-exclude-params",
            dest="qase_exclude_params",
            help="Exclude params from test results"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-execution-plan-path",
            dest="qase_execution_plan_path",
            help="Path to file with execution plan"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-testops-project",
            dest="qase_testops_project",
            help="Project code in Qase TestOps"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-testops-api-token",
            dest="qase_testops_api_token",
            help="API token for Qase TestOps"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-testops-api-host",
            dest="qase_testops_api_host",
            help="API host for Qase TestOps"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-testops-plan-id",
            dest="qase_testops_plan_id",
            help="Test Plan ID in Qase TestOps"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-testops-run-id",
            dest="qase_testops_run_id",
            help="Test Run ID in Qase TestOps"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-testops-run-tags",
            dest="qase_testops_run_tags",
            help="Test Run tags in Qase TestOps"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-testops-run-title",
            "qase_testops_run_title",
            help="Define title for autocreated Test Run. If not set, will be used autogenerated title"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-testops-run-description",
            dest="qase_testops_run_description",
            help="Define description for autocreated Test Run"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-testops-run-complete",
            dest="qase_testops_run_complete",
            type="bool",
            help="Complete run after tests execution"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-testops-run-external-link-type",
            dest="qase_testops_run_external_link_type",
            help="External link type for test run: `jiraCloud` or `jiraServer`"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-testops-run-external-link-url",
            dest="qase_testops_run_external_link_url",
            help="External link URL or identifier for test run"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-testops-batch-size",
            dest="qase_testops_batch_size",
            type="int",
            help="Define batch size of results. Default: 200."
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-testops-defect",
            dest="qase_testops_defect",
            type="bool",
            help="Create defect if test failed"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-testops-configurations-values",
            dest="qase_testops_configurations_values",
            help="Configurations for test run. Format: 'group1=value1,group2=value2'"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-testops-configurations-create-if-not-exists",
            dest="qase_testops_configurations_create_if_not_exists",
            type="bool",
            help="Create configurations if they don't exist. Default: false"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-testops-status-filter",
            dest="qase_testops_status_filter",
            help="Filter results by status. Format: 'passed,failed,skipped'"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-testops-show-public-report-link",
            dest="qase_testops_show_public_report_link",
            type="bool",
            help="Enable automatic generation of public report link after test run completion"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-status-mapping",
            dest="qase_status_mapping",
            help="Map test result statuses. Format: 'invalid=failed,skipped=passed'"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-report-driver",
            dest="qase_report_driver",
            help="Define report driver: `local`. More options coming soon."
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-root-suite",
            dest="qase_root_suite",
            help="Define root suite for tests"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-report-connection-local-path",
            dest="qase_report_connection_local_path",
            help="Define local path for report directory"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-report-connection-local-format",
            dest="qase_report_connection_local_format",
            help="Define local format for report directory: `json` or `jsonp`"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-profilers",
            dest="qase_profilers",
            help="Profilers to use for tests. Available: `network`, `db`, `sleep`"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-pytest-capture-logs",
            dest="qase_pytest_capture_logs",
            type="bool",
            help="Capture logs from pytest"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-pytest-xfail-status-xfail",
            dest="qase_pytest_xfail_status",
            help="Define xfail status for failed tests. Default: `skipped`"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-pytest-xfail-status-xpass",
            dest="qase_pytest_xpass_status",
            help="Define xpass status for passed tests. Default: `passed`"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-logging-console",
            dest="qase_logging_console",
            type="bool",
            help="Enable/disable console logging output"
        )

        QasePytestOptions.add_option(
            parser,
            group,
            "--qase-logging-file",
            dest="qase_logging_file",
            type="bool",
            help="Enable/disable file logging output"
        )

    @staticmethod
    def add_option(parser, group, option, dest, default=None, type=None, **kwargs):
        # We are going to add options that were not added before through the manager
        try:
            group.addoption(option, dest=dest, default=default, action="store", **kwargs)
        except ValueError:
            pass
